import codecs
import json

import siphash


class AuthMethod(object):
    SIMPLE_AUTH = "sa"
    TIMESTAMP_AUTH = "ta"
    COUNTER_AUTH = "ca"
    DATA_AUTH = "da"
    RECURSIVE_DATA_AUTH = "ra"


class OpenPAYGOMetricsShared(object):
    CONDENSED_KEY_NAMES = {
        # Request
        "serial_number": "sn",
        "timestamp": "ts",
        "auth": "a",
        "request_count": "rc",
        "data_collection_timestamp": "dtc",
        "data_format_id": "df",
        "data_format": "dfo",
        "data": "d",
        "historical_data": "hd",
        "accessories": "acc",
        # Response
        "token_list": "tkl",
        "active_until_timestamp": "auts",
        "active_seconds_left": "asl",
        "settings": "st",
        "extra_data": "ed",
        # Data
        "token_count": "tc",
        "active_until_timestamp_requested": "autsr",
        "active_seconds_left_requested": "aslr",
    }

    @classmethod
    def convert_dict_keys_to_condensed(cls, simple_dict):
        return cls.convert_dict_keys(simple_dict, cls.CONDENSED_KEY_NAMES)

    @classmethod
    def convert_dict_keys_to_simple(cls, condensed_dict):
        revert_keys = {v: k for k, v in cls.CONDENSED_KEY_NAMES.items()}
        return cls.convert_dict_keys(condensed_dict, revert_keys)

    @classmethod
    def convert_dict_keys(cls, origin_dict, key_map):
        condensed_dict = {}
        for key in origin_dict:
            if key in key_map:
                condensed_dict[key_map[key]] = origin_dict[key]
            else:
                condensed_dict[key] = origin_dict[key]
        return condensed_dict

    @classmethod
    def remove_trailing_empty_elements(cls, list_with_empty):
        while list_with_empty and list_with_empty[-1] is None:
            list_with_empty.pop()
        return list_with_empty

    @classmethod
    def convert_to_metrics_json(cls, data):
        return json.dumps(data, separators=(",", ":"))

    @classmethod
    def generate_response_signature_from_data(
        cls, data, secret_key, serial_number, timestamp=None, request_count=None
    ):
        payload = serial_number
        if timestamp:
            payload += str(timestamp)
        if request_count:
            payload += str(request_count)
        if data.get("active_until_timestamp"):
            payload += str(data.get("active_until_timestamp"))
        if data.get("active_seconds_left"):
            payload += str(data.get("active_seconds_left"))
        if data.get("token_list"):
            payload += cls.convert_to_metrics_json(data.get("token_list"))
        if data.get("settings"):
            payload += cls.convert_to_metrics_json(data.get("settings"))
        if data.get("extra_data"):
            payload += cls.convert_to_metrics_json(data.get("extra_data"))
        return AuthMethod.DATA_AUTH + cls.generate_hash_string(payload, secret_key)

    @classmethod
    def generate_request_signature_from_data(cls, data, auth_method, secret_key):
        if auth_method == AuthMethod.SIMPLE_AUTH:
            signature = cls.generate_hash_string(data.get("serial_number"), secret_key)
        elif auth_method == AuthMethod.TIMESTAMP_AUTH:
            if not data.get("timestamp", None):
                raise ValueError("Timestamp is required for Timestamp Auth")
            signature = cls.generate_hash_string(
                data.get("serial_number") + str(data.get("timestamp")), secret_key
            )
        elif auth_method == AuthMethod.COUNTER_AUTH:
            if not data.get("request_count", None):
                raise ValueError("Request Count is required for Counter Auth")
            signature = cls.generate_hash_string(
                data.get("serial_number") + str(data.get("request_count")), secret_key
            )
        elif auth_method == AuthMethod.DATA_AUTH:
            payload = data.get("serial_number")
            if data.get("timestamp"):
                payload += str(data.get("timestamp"))
            if data.get("request_count"):
                payload += str(data.get("request_count"))
            if data.get("data"):
                payload += cls.convert_to_metrics_json(data.get("data", []))
            if data.get("historical_data"):
                payload += cls.convert_to_metrics_json(data.get("historical_data", []))
            signature = cls.generate_hash_string(payload, secret_key)
        elif auth_method == AuthMethod.RECURSIVE_DATA_AUTH:
            payload = data.get("serial_number")
            payload = cls.generate_hash_string(payload, secret_key)
            if data.get("timestamp"):
                payload = cls.generate_hash_string(
                    payload + str(data.get("timestamp")), secret_key
                )
            if data.get("request_count"):
                payload = cls.generate_hash_string(
                    payload + str(data.get("request_count")), secret_key
                )
            payload = cls.generate_hash_string(
                payload + cls.convert_to_metrics_json(data.get("data", [])), secret_key
            )
            for time_step_data in data.get("historical_data", []):
                payload = cls.generate_hash_string(
                    payload + cls.convert_to_metrics_json(time_step_data), secret_key
                )
            signature = payload
        else:
            raise ValueError("Invalid Authentication Method")
        return auth_method + signature

    @classmethod
    def generate_hash_string(cls, input_string, secret_key):
        key = cls.load_secret_key_from_hex(secret_key)
        hash = siphash.SipHash_2_4(key, codecs.encode(input_string, "utf-8")).hash()
        hash_string = "{:x}".format(hash)
        return hash_string

    @classmethod
    def load_secret_key_from_hex(cls, secret_key):
        try:
            return codecs.decode(secret_key, "hex")
        except Exception:
            raise ValueError(
                "The secret key provided is not correctly formatted, it should be 32 "
                "hexadecimal characters. "
            )



import copy

from .metrics_shared import OpenPAYGOMetricsShared


class MetricsRequestHandler(object):
    def __init__(
        self, serial_number, data_format=None, secret_key=None, auth_method=None
    ):
        self.secret_key = secret_key
        self.auth_method = auth_method
        self.request_dict = {
            "serial_number": serial_number,
        }
        self.data_format = data_format
        if self.data_format:
            if self.data_format.get("id"):
                self.request_dict["data_format_id"] = data_format.get("id")
            else:
                self.request_dict["data_format"] = data_format
        self.data = {}
        self.historical_data = {}

    def set_request_count(self, request_count):
        self.request_dict["request_count"] = request_count

    def set_timestamp(self, timestamp):
        self.request_dict["timestamp"] = timestamp

    def set_data(self, data):
        self.data = data

    def set_historical_data(self, historical_data):
        if not self.data_format.get("historical_data_interval"):
            for time_step in historical_data:
                if not time_step.get("timestamp"):
                    raise ValueError(
                        "Historical Data objects must have a time stamp if no "
                        "historical_data_interval is defined."
                    )
        self.historical_data = historical_data

    def get_simple_request_payload(self):
        payload = self.get_simple_request_dict()
        return OpenPAYGOMetricsShared.convert_to_metrics_json(payload)

    def get_simple_request_dict(self):
        simple_request = self.request_dict
        simple_request["data"] = self.data
        simple_request["historical_data"] = self.historical_data
        # We prepare the auth
        if self.auth_method:
            simple_request["auth"] = (
                OpenPAYGOMetricsShared.generate_request_signature_from_data(
                    simple_request, self.auth_method, self.secret_key
                )
            )
        return simple_request

    def get_condensed_request_payload(self):
        payload = self.get_condensed_request_dict()
        return OpenPAYGOMetricsShared.convert_to_metrics_json(payload)

    def get_condensed_request_dict(self):
        if not self.data_format:
            raise ValueError("No Data Format provided for condensed request")
        data_order = self.data_format.get("data_order")
        if self.data and not data_order:
            raise ValueError("Data Format does not contain data_order")
        historical_data_order = self.data_format.get("historical_data_order")
        if self.historical_data and not historical_data_order:
            raise ValueError("Data Format does not contain historical_data_order")
        condensed_request = copy.deepcopy(self.request_dict)
        condensed_request["data"] = []
        condensed_request["historical_data"] = []
        # We add the data
        data_copy = copy.deepcopy(self.data)
        for var in data_order:
            condensed_request["data"].append(
                data_copy.pop(var) if var in data_copy else None
            )
        if len(data_copy) > 0:
            raise ValueError(
                "Additional variables not present in the data format: " + str(data_copy)
            )
        condensed_request["data"] = (
            OpenPAYGOMetricsShared.remove_trailing_empty_elements(
                condensed_request["data"]
            )
        )
        # We add the historical data
        historical_data_copy = copy.deepcopy(self.historical_data)
        for time_step in historical_data_copy:
            time_step_data = []
            for var in historical_data_order:
                time_step_data.append(time_step.pop(var) if var in time_step else None)
            if len(time_step) > 0:
                raise ValueError(
                    "Additional variables not present in the historical data format: "
                    + str(time_step)
                )
            time_step_data = OpenPAYGOMetricsShared.remove_trailing_empty_elements(
                time_step_data
            )
            condensed_request["historical_data"].append(time_step_data)
        # We prepare the auth
        if self.auth_method:
            condensed_request["auth"] = (
                OpenPAYGOMetricsShared.generate_request_signature_from_data(
                    condensed_request, self.auth_method, self.secret_key
                )
            )
        # We replace the key names by the condensed ones
        condensed_request = OpenPAYGOMetricsShared.convert_dict_keys_to_condensed(
            condensed_request
        )
        return condensed_request



import json

import pytest

from openpaygo import OpenPAYGOTokenEncoder, TokenType

with open("test/test_tokens.jsonl") as f:
    sample_data = [json.loads(line) for line in f]


@pytest.fixture
def encoder():
    return OpenPAYGOTokenEncoder()


@pytest.fixture
def token_type_lookup():
    return {
        "ADD_TIME": TokenType.ADD_TIME,
        "SET_TIME": TokenType.SET_TIME,
        "DISABLE_PAYG": TokenType.DISABLE_PAYG,
        "COUNTER_SYNC": TokenType.COUNTER_SYNC,
    }


@pytest.mark.parametrize("data", sample_data)
def test_generate_token(encoder, token_type_lookup, data):
    new_count, final_token = encoder.generate_token(
        secret_key=data["key"],
        count=data["count"],
        value=data["value_raw"],
        token_type=token_type_lookup[data["token_type"]],
        starting_code=data["starting_code"],
        restricted_digit_set=data["restricted_digit_set"],
        extended_token=data["extended_token"],
    )

    assert new_count == data["new_count"]
    assert final_token == data["token"]



import json

from openpaygo import OpenPAYGOTokenEncoder, TokenType

encoder = OpenPAYGOTokenEncoder()

token_types = [
    "ADD_TIME",
    "SET_TIME",
    "DISABLE_PAYG",
    "COUNTER_SYNC",
]

token_add_set_time_values = [
    1,
    2,
    5,
    995,  # highest allowed value without special meaning
]
token_type_lookup = {
    "ADD_TIME": TokenType.ADD_TIME,
    "SET_TIME": TokenType.SET_TIME,
    "DISABLE_PAYG": TokenType.DISABLE_PAYG,
    "COUNTER_SYNC": TokenType.COUNTER_SYNC,
}

restricted_digit_set_values = [False, True]

extended_token_values = [False, True]

all_cases = [
    {
        "token_type": token_type,
        "token_value": token_value,
        "restricted_digit_set": restricted_digit_set,
        "extended_token": extended_token,
    }
    for token_type in token_types
    for token_value in (
        token_add_set_time_values if token_type in ["ADD_TIME", "SET_TIME"] else [None]
    )
    for restricted_digit_set in restricted_digit_set_values
    for extended_token in extended_token_values
]


def array_of_dicts_to_jsonl(array, filename):
    with open(filename, "w") as outfile:
        for item in array:
            json.dump(item, outfile)
            outfile.write("\n")


if __name__ == "__main__":
    device_data1 = {
        "serial_number": "TEST220000001",
        "starting_code": 516959010,
        "key": "bc41ec9530f6dac86b1a29ab82edc5fb",
        "count": 1,
    }
    device_data2 = {
        "serial_number": "TEST240000002",
        "starting_code": 432435255,
        "key": "dac86b1a29ab82edc5fbbc41ec9530f6",
        "count": 10,
    }

    test_cases = []

    # DEVICE 1
    for t in all_cases:
        new_count, final_token = encoder.generate_token(
            secret_key=device_data1["key"],
            count=device_data1["count"],
            value=t["token_value"],
            token_type=token_type_lookup[t["token_type"]],
            starting_code=device_data1["starting_code"],
            restricted_digit_set=t["restricted_digit_set"],
            extended_token=t["extended_token"],
        )

        result = {
            **device_data1,
            **{
                "token_type": t["token_type"],
                "value_raw": t["token_value"],
                "restricted_digit_set": t["restricted_digit_set"],
                "extended_token": t["extended_token"],
            },
            **{"new_count": new_count, "token": final_token.replace(" ", "")},
        }
        test_cases.append(result)

    # DEVICE 2
    for t in all_cases:
        new_count, final_token = encoder.generate_token(
            secret_key=device_data2["key"],
            count=device_data2["count"],
            value=t["token_value"],
            token_type=token_type_lookup[t["token_type"]],
            starting_code=device_data2["starting_code"],
            restricted_digit_set=t["restricted_digit_set"],
            extended_token=t["extended_token"],
        )

        result = {
            **device_data2,
            **{
                "token_type": t["token_type"],
                "value_raw": t["token_value"],
                "restricted_digit_set": t["restricted_digit_set"],
                "extended_token": t["extended_token"],
            },
            **{"new_count": new_count, "token": final_token.replace(" ", "")},
        }
        test_cases.append(result)

    array_of_dicts_to_jsonl(test_cases, "test/test_tokens.jsonl")



from .metrics_request import MetricsRequestHandler
from .metrics_response import MetricsResponseHandler
from .metrics_shared import AuthMethod
from .token_decode import OpenPAYGOTokenDecoder
from .token_encode import OpenPAYGOTokenEncoder
from .token_shared import TokenType


def generate_token(**kwargs):
    return OpenPAYGOTokenEncoder.generate_token(**kwargs)


def decode_token(**kwargs):
    return OpenPAYGOTokenDecoder.decode_token(**kwargs)


__all__ = [
    MetricsRequestHandler,
    MetricsResponseHandler,
    AuthMethod,
    OpenPAYGOTokenDecoder,
    OpenPAYGOTokenEncoder,
    TokenType,
]



from .token_shared import OpenPAYGOTokenShared, TokenType
from .token_shared_extended import OpenPAYGOTokenSharedExtended


class OpenPAYGOTokenEncoder(object):
    @classmethod
    def generate_token(
        cls,
        secret_key,
        count,
        value=None,
        token_type=TokenType.ADD_TIME,
        starting_code=None,
        value_divider=1,
        restricted_digit_set=False,
        extended_token=False,
    ):
        secret_key = OpenPAYGOTokenShared.load_secret_key_from_hex(secret_key)
        if not starting_code:
            # We generate the starting code from the key if not provided
            starting_code = OpenPAYGOTokenShared.generate_starting_code(secret_key)
        if token_type in [TokenType.ADD_TIME, TokenType.SET_TIME]:
            value = int(round(value * value_divider, 0))
            if not extended_token:
                max_value = OpenPAYGOTokenShared.MAX_ACTIVATION_VALUE
            else:
                max_value = OpenPAYGOTokenSharedExtended.MAX_ACTIVATION_VALUE
            if value > max_value:
                raise ValueError("The value provided is too high.")
        elif value:
            raise ValueError("A value is not allowed for this token type.")
        else:
            if token_type == TokenType.DISABLE_PAYG:
                value = OpenPAYGOTokenShared.PAYG_DISABLE_VALUE
            elif token_type == TokenType.COUNTER_SYNC:
                value = OpenPAYGOTokenShared.COUNTER_SYNC_VALUE
            else:
                raise ValueError("The token type provided is not supported.")
        if extended_token:
            return cls.generate_extended_token(
                starting_code,
                secret_key,
                value,
                count,
                token_type,
                restricted_digit_set,
            )
        else:
            return cls.generate_standard_token(
                starting_code,
                secret_key,
                value,
                count,
                token_type,
                restricted_digit_set,
            )

    @classmethod
    def generate_standard_token(
        cls,
        starting_code=None,
        key=None,
        value=None,
        count=None,
        mode=TokenType.ADD_TIME,
        restricted_digit_set=False,
    ):
        # We get the first 3 digits with encoded value
        starting_code_base = OpenPAYGOTokenShared.get_token_base(starting_code)
        token_base = cls._encode_base(starting_code_base, value)
        current_token = OpenPAYGOTokenShared.put_base_in_token(
            starting_code, token_base
        )
        new_count = cls._get_new_count(count, mode)
        for xn in range(0, new_count):
            current_token = OpenPAYGOTokenShared.generate_next_token(current_token, key)
        final_token = OpenPAYGOTokenShared.put_base_in_token(current_token, token_base)
        if restricted_digit_set:
            final_token = OpenPAYGOTokenShared.convert_to_4_digit_token(final_token)
            final_token = "{:015d}".format(final_token)
        else:
            final_token = "{:09d}".format(final_token)
        return new_count, final_token

    @classmethod
    def _encode_base(cls, base, number):
        if number + base > 999:
            return number + base - 1000
        else:
            return number + base

    @classmethod
    def generate_extended_token(
        cls,
        starting_code,
        key,
        value,
        count,
        mode=TokenType.ADD_TIME,
        restricted_digit_set=False,
    ):
        starting_code_base = OpenPAYGOTokenSharedExtended.get_token_base(starting_code)
        token_base = cls._encode_base_extended(starting_code_base, value)
        current_token = OpenPAYGOTokenSharedExtended.put_base_in_token(
            starting_code, token_base
        )
        new_count = cls._get_new_count(count, mode)
        for xn in range(0, new_count):
            current_token = OpenPAYGOTokenSharedExtended.generate_next_token(
                current_token, key
            )
        final_token = OpenPAYGOTokenSharedExtended.put_base_in_token(
            current_token, token_base
        )
        if restricted_digit_set:
            final_token = OpenPAYGOTokenSharedExtended.convert_to_4_digit_token(
                final_token
            )
            final_token = "{:020d}".format(final_token)
        else:
            final_token = "{:012d}".format(final_token)
        return new_count, final_token

    @classmethod
    def _encode_base_extended(cls, base, number):
        if number + base > 999999:
            return number + base - 1000000
        else:
            return number + base

    @classmethod
    def _get_new_count(cls, count, mode):
        current_count_odd = count % 2
        if mode in [TokenType.SET_TIME, TokenType.DISABLE_PAYG, TokenType.COUNTER_SYNC]:
            if (
                current_count_odd
            ):  # Odd numbers are for Set Time, Disable PAYG or Counter Sync
                new_count = count + 2
            else:
                new_count = count + 1
        else:
            if current_count_odd:  # Even numbers are for Add Time
                new_count = count + 1
            else:
                new_count = count + 2
        return new_count



import copy
import json
from datetime import datetime, timedelta

from .metrics_shared import OpenPAYGOMetricsShared


class MetricsResponseHandler(object):
    def __init__(
        self,
        received_metrics,
        data_format=None,
        secret_key=None,
        last_request_count=None,
        last_request_timestamp=None,
    ):
        self.received_metrics = received_metrics
        self.request_dict = json.loads(received_metrics)
        # We convert the base variable names to simple
        self.request_dict = OpenPAYGOMetricsShared.convert_dict_keys_to_simple(
            self.request_dict
        )
        # We add the reception timestamp if not timestamp was provided
        self.request_timestamp = self.request_dict.get("timestamp")
        if not self.request_dict.get("timestamp"):
            self.timestamp = int(datetime.now().timestamp())
        else:
            self.timestamp = self.request_dict.get("timestamp")
        self.response_dict = {}
        self.secret_key = secret_key
        self.data_format = data_format
        self.last_request_count = last_request_count
        self.last_request_timestamp = last_request_timestamp
        if not self.data_format and self.request_dict.get("data_format"):
            self.data_format = self.request_dict.get("data_format")

    def get_device_serial(self):
        return self.request_dict.get("serial_number")

    def get_data_format_id(self):
        return self.request_dict.get("data_format_id")

    def data_format_available(self):
        return self.data_format is not None

    def set_device_parameters(
        self,
        secret_key=None,
        data_format=None,
        last_request_count=None,
        last_request_timestamp=None,
    ):
        if secret_key:
            self.secret_key = secret_key
        if data_format:
            self.data_format = data_format
        if last_request_count:
            self.last_request_count = last_request_count
        if last_request_timestamp:
            self.last_request_timestamp = last_request_timestamp

    def is_auth_valid(self):
        auth_string = self.request_dict.get("auth", None)
        if not auth_string:
            return False
        elif not self.secret_key:
            raise ValueError("Secret key is required to check the auth.")
        self.auth_method = auth_string[:2]
        new_signature = OpenPAYGOMetricsShared.generate_request_signature_from_data(
            self.request_dict, self.auth_method, self.secret_key
        )
        if auth_string == new_signature:
            request_count = self.get_request_count()
            if (
                request_count
                and self.last_request_count
                and request_count <= self.last_request_count
            ):
                return False
            timestamp = self.get_request_timestamp()
            if (
                timestamp
                and self.last_request_timestamp
                and timestamp <= self.last_request_timestamp
            ):
                return False
            # Either the request count or timestamp is required
            if request_count or timestamp:
                return True
        return False

    def get_simple_metrics(self):
        # We start the process by making a copy of the dict to work with
        simple_dict = copy.deepcopy(self.request_dict)
        simple_dict.pop("auth") if "auth" in simple_dict else None  # We remove the auth
        # We process the data and replace it
        simple_dict["data"] = self._get_simple_data()
        # We process the historical data
        simple_dict["historical_data"] = self._get_simple_historical_data()
        # We fill in the timestamps for each time step
        simple_dict["historical_data"] = self._fill_timestamp_in_historical_data(
            simple_dict["historical_data"]
        )
        return simple_dict

    def get_data_timestamp(self):
        return self.request_dict.get("data_collection_timestamp", self.timestamp)

    def get_request_timestamp(self):
        return self.request_timestamp

    def get_request_count(self):
        return self.request_dict.get("request_count")

    def get_token_count(self):
        data = self._get_simple_data()
        return data.get("token_count")

    def expects_token_answer(self):
        return self.get_token_count() is not None

    def add_tokens_to_answer(self, token_list):
        self.response_dict["token_list"] = token_list

    def expects_time_answer(self):
        data = self._get_simple_data()
        if data.get("active_until_timestamp_requested", False) or data.get(
            "active_seconds_left_requested", False
        ):
            return True
        return False

    def add_time_to_answer(self, target_datetime):
        data = self._get_simple_data()
        if data.get("active_until_timestamp_requested", False):
            target_timestamp = 0
            if target_datetime:
                if target_datetime.year > 1970:
                    target_timestamp = target_datetime.timestamp()
            self.response_dict["active_until_timestamp"] = target_timestamp
        elif data.get("active_seconds_left_requested", False):
            seconds_left = (
                (datetime.now() - target_datetime).total_seconds()
                if target_datetime
                else 0
            )
            self.response_dict["active_seconds_left"] = (
                seconds_left if seconds_left > 0 else 0
            )
        else:
            raise ValueError("No time requested")

    def add_new_base_url_to_answer(self, new_base_url):
        self.add_settings_to_answer({"base_url": new_base_url})

    def add_settings_to_answer(self, settings_dict):
        if not self.response_dict.get("settings"):
            self.response_dict["settings"] = {}
        self.response_dict["settings"].update(settings_dict)

    def add_extra_data_to_answer(self, extra_data_dict):
        if not self.response_dict.get("extra_data"):
            self.response_dict["extra_data"] = {}
        self.response_dict["extra_data"].update(extra_data_dict)

    def get_answer_payload(self):
        payload = self.get_answer_dict()
        return OpenPAYGOMetricsShared.convert_to_metrics_json(payload)

    def get_answer_dict(self):
        # If there is not data format, we just return the full response
        condensed_answer = copy.deepcopy(self.response_dict)
        if self.secret_key:
            condensed_answer["auth"] = (
                OpenPAYGOMetricsShared.generate_response_signature_from_data(
                    serial_number=self.request_dict.get("serial_number"),
                    request_count=self.request_dict.get("request_count"),
                    data=condensed_answer,
                    timestamp=self.request_dict.get("timestamp"),
                    secret_key=self.secret_key,
                )
            )
        return OpenPAYGOMetricsShared.convert_dict_keys_to_condensed(condensed_answer)

    def _get_simple_data(self):
        data = copy.deepcopy(self.request_dict.get("data"))
        # If no data or not condensed in list, we just return it
        if not data:
            return {}
        if not isinstance(data, list):
            return data
        data_order = self.data_format.get("data_order")
        if not data_order:
            raise ValueError("Data Format does not contain data_order")
        clean_data = {}
        data_len = len(data)
        for idx, var in enumerate(data_order):
            clean_data[var] = data[idx] if idx < data_len else None
        data = data[data_len:]
        if len(data) > 0:
            raise ValueError(
                "Additional variables not present in the data format: " + str(data)
            )
        return OpenPAYGOMetricsShared.convert_dict_keys_to_simple(clean_data)

    def _get_simple_historical_data(self):
        historical_data = copy.deepcopy(self.request_dict.get("historical_data"))
        if not historical_data:
            return []
        historical_data_order = self.data_format.get("historical_data_order")
        clean_historical_data = []
        for time_step in historical_data:
            time_step_data = {}
            if isinstance(time_step, list):
                if not historical_data_order:
                    raise ValueError(
                        "Data Format does not contain historical_data_order"
                    )
                timse_step_len = len(time_step)
                for idx, var in enumerate(historical_data_order):
                    if idx < timse_step_len:
                        time_step_data[var] = time_step[idx]
                time_step = time_step[timse_step_len:]
                if len(time_step) > 0:
                    raise ValueError(
                        "Additional variables not present in the historical data format"
                        ": " + str(time_step)
                    )
            elif isinstance(time_step, dict):
                for key in time_step:
                    if key.isdigit() and int(key) < len(historical_data_order):
                        time_step_data[historical_data_order[int(key)]] = time_step[key]
                    else:
                        time_step_data[key] = time_step[key]
            else:
                raise ValueError("Invalid historical data step type: " + str(time_step))
            clean_historical_data.append(time_step_data)
        return clean_historical_data

    def _fill_timestamp_in_historical_data(self, historical_data):
        last_timestamp = datetime.fromtimestamp(self.get_data_timestamp())
        for idx, time_step in enumerate(historical_data):
            if time_step.get("relative_time") is not None:
                last_timestamp = last_timestamp + timedelta(
                    seconds=int(time_step.get("relative_time"))
                )
                historical_data[idx]["timestamp"] = int(last_timestamp.timestamp())
                del historical_data[idx]["relative_time"]
            elif time_step.get("timestamp"):
                last_timestamp = datetime.fromtimestamp(time_step.get("timestamp"))
            else:
                if idx != 0:
                    last_timestamp = last_timestamp + timedelta(
                        seconds=int(self.data_format.get("historical_data_interval"))
                    )
                historical_data[idx]["timestamp"] = int(last_timestamp.timestamp())
        return historical_data
